from .common_webscraper_functions import get_value_from_element, does_html_object_exist
import logging

class FieldParserFactory:

    def __init__(self, field_dict):
        self.field_dict = field_dict

    def create(self):
        if self.field_dict['parse_type'] == 'static':
            return StaticFieldParser(self.field_dict['field_name'], self.field_dict['static_value'])
        elif self.field_dict['parse_type'] == 'input_dict':
            return InputDictFieldParser(self.field_dict['field_name'], self.field_dict['dict_key'])
        elif self.field_dict['parse_type'] == 'dynamic':
            if 'html_object' in self.field_dict.keys():
                return ChildElementParser(self.field_dict['field_name'], self.field_dict)
            return DynamicFieldParser(self.field_dict['field_name'], self.field_dict)
        raise Exception("No Match for parse type " + self.field_dict['parse_type'])

class FieldParser:

    def __init__(self, field_name):
        self.field_name = field_name

    def parse(self, html_object, data_dict, webscraperObject):
        pass

class InputDictFieldParser(FieldParser): 

    def __init__(self, field_name, input_dict_field):
        super().__init__(field_name)
        self.input_dict_field = input_dict_field

    def parse(self, html_object, data_dict, webscraperObject):
        return data_dict[self.input_dict_field]

class StaticFieldParser(FieldParser):

    def __init__(self, field_name, static_value):
        super().__init__(field_name)
        self.static_value = static_value


    def parse(self, html_object, data_dict, webscraperObject):
        return self.static_value
    
class DynamicFieldParser(FieldParser):

    def __init__(self, field_name, field_dict):
        super().__init__(field_name)
        self.field_dict = field_dict
        self.function = get_value_from_element(field_dict)

    def parse(self, html_object, data_dict, webscraperObject):
        logging.info("[WEBSCRAPER] [DynamicFieldParser] [parse] html_object: " + str(html_object) + " data_dict: " + str(data_dict))
        return_value = self.function(html_object)
        logging.info("[WEBSCRAPER] [DynamicFieldParser] [parse] return_value: " + str(return_value))
        if 'object_name' in self.field_dict.keys():
            try:
                return webscraperObject.databaseObject.tables[self.field_dict['object_name']].get_primary_key_by_search_dict({'url': return_value})
            except:
                logging.info("[WEBSCRAPER] [DynamicFieldParser] Unable to find object match for: " + str(return_value))
                return webscraperObject.get_webscraper_object_with_name(self.field_dict['object_name']).create_from_page(return_value, webscraperObject)
        return return_value
    
class ChildElementParser(FieldParser):

    def __init__(self, field_name, field_dict):
        super().__init__(field_name)
        self.function = does_html_object_exist(field_dict)

    def parse(self, html_object, data_dict, webscraperObject):
        return self.function(html_object)


from bs4 import BeautifulSoup
from selenium import webdriver
import selenium
import logging

def get_element(field_dict):
    element_type = None
    attributes = {}
    if 'element_type' in field_dict.keys():
        element_type = field_dict['element_type']
    if 'attributes' in field_dict.keys():
        attributes = field_dict['attributes']
    if 'id' in field_dict.keys():
        def return_function(html_object):
            return html_object.find(id=field_dict['id'])
        return return_function
    if element_type is None:
        def return_function(html_object):
            return html_object.find(attrs=attributes)
        return return_function
    def return_function(html_object):
        return html_object.find(element_type, attrs=attributes)
    return return_function

def get_children_element(field_dict):
    element_type = None
    attributes = {}
    if 'element_type' in field_dict.keys():
        element_type = field_dict['element_type']
    if 'attributes' in field_dict.keys():
        attributes = field_dict['attributes']
    if 'id' in field_dict.keys():
        def return_function(html_object):
            return html_object.find_all(id=field_dict['id'])
        return return_function
    if element_type is None:
        def return_function(html_object):
            return html_object.find_all(attrs=attributes)
        return return_function
    def return_function(html_object):
        return html_object.find_all(element_type, attrs=attributes)
    return return_function

def get_rough_value_from_element(field_dict):
    logging.info("[GET_ROUGH_VALUE_FROM_ELEMENT] " + str(field_dict))
    if field_dict['html_field_value'] == 'url':
        def return_function(html_object):
            return get_element(field_dict)(html_object).find('a')['href']
        return return_function
    elif field_dict['html_field_value'] == 'text':
        def return_function(html_object):
            return get_element(field_dict)(html_object).text
        return return_function
    else:
        raise Exception(f"No match for html field value {field_dict['html_field_value']}")

def get_value_from_element(field_dict):
    def return_function(html_object):
        return_value = get_rough_value_from_element(field_dict)(html_object)
        if 'remove_strings' in field_dict.keys():
            for remove_string in field_dict['remove_strings']:
                return_value = return_value.replace(remove_string, '')
        return return_value
    return return_function

def does_html_object_exist(field_dict):
    def return_function(html_object):
        if get_element(field_dict)(html_object).find(field_dict['html_object']) is not None:
            return 1
        return 0
    return return_function

def get_tr_of_table_with_id(table_id):
    def return_function(soup):
        return soup.find(id=table_id).find("tbody").find_all("tr")
    return return_function

def get_tr_of_stats_table():
    def return_function(soup):
        return soup.find(attrs={'class': 'stats_table'}).find('tbody').find_all('tr')
    return return_function

def row_has_link(html_object):
    return html_object.find("a") is not None

def true(html_object):
    return True

def fetch_soup_from_page(url):
    print(url)
    while True:
        try:
            options = webdriver.ChromeOptions()
            options.add_argument('headless')
            options.add_argument('window-size=1200x600')
            driver = webdriver.Chrome(options=options)
            driver.get(url)
            page = driver.page_source
            driver.quit()
            soup = BeautifulSoup(page, 'html.parser')
            return soup
        except selenium.common.exceptions.TimeoutException:
            logging.warn("Timeout")
            print("Failing url: url")
            print("Timed out loading page, trying again")
        except selenium.common.exceptions.WebDriverException:
            logging.warn("WebdriverException")
            print("failing url: " + str(url))
            print("Web Driver Error, trying again")



class URLGenerator:

    def __init__(self, base_url, iterator=None):
        self.iterator = iterator
        self.base_url = base_url

    def generate_urls(self, webscraper_object_collection):
        if self.iterator is None:
            return [{'data_dict': {}, 'url': self.base_url}]
        return self.iterator.generate_iterations(self.base_url, webscraper_object_collection)

class URLIteratorFactory():

    def __init__(self, url_iterator_dict):
        self.url_iterator_dict = url_iterator_dict

    def create(self):
        if 'hardcoded' in self.url_iterator_dict.keys():
            return HardcodedURLIterator(self.url_iterator_dict['hardcoded'])
        elif 'object' in self.url_iterator_dict.keys():
            return ObjectURLIterator(self.url_iterator_dict['object'])
        else:
            raise Exception("No match for url iterator type")

class URLIterator():

    def __init__(self):
        pass

    def generate_iterations(self, base_url, webscraper_object_collection):
        pass

class HardcodedURLIterator(URLIterator):

    def __init__(self, hardcoded_strings):
        self.hardcoded_strings = hardcoded_strings
        super().__init__()
    
    def generate_iterations(self, base_url, webscraper_object_collection):
        return_array = []
        for hardcoded_string in self.hardcoded_strings:
            return_array.append({'data_dict': {}, 'url': base_url.replace('{replace}', hardcoded_string)})
        return return_array

class ObjectURLIterator(URLIterator):

    def __init__(self, object_name):
        self.object_name = object_name
        super().__init__()

    def generate_iterations(self, base_url, webscraper_object_collection):
        return_array = []
        for data_dict in webscraper_object_collection.databaseObject.tables[self.object_name].data:
            return_array.append({'url': base_url.replace('{replace}', data_dict['url']), 'data_dict': data_dict})
        return return_array
    
class URLGeneratorFactory:

    def __init__(self, url_dict):
        self.url_dict = url_dict

    def get_url_generator(self):
        if 'iterator' in self.url_dict.keys():
            return URLGenerator(self.url_dict['base_url'], URLIteratorFactory(self.url_dict['iterator']).create())
        else:
            return URLGenerator(self.url_dict['base_url'])